/*
 *      MIPL Diagnostics utility
 *
 *      Authors:
 *      Antti Tuominen          <ajtuomin@tml.hut.fi>
 *
 *      $Id: mipdiag.c,v 1.6 2001/02/22 14:39:48 antti Exp $
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 *      Changes:
 *      Alexandru Petrescu,
 *      Miguel Catalina-Gallego   :    Support for configuration of keys 
 *                                     via /proc/sys/... 
 */


#define ERR_UID       10
#define ERR_ARG       11
#define ERR_NO_PROC   15
#define ERR_NO_SYSCTL 16
#define EINVDIGIT       1
#define ESRCTOOLONG     2
#define ESRCTOOSHORT    3

#define UID0 0

#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <inttypes.h>
#include <netinet/in.h>
#include <ctype.h>

static char base64_digits [] =
"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz- ";

void usage();
void version();

void print_access();
void print_bcache();
void print_bulist();
void print_stats();

void access_do_one(char policy, char *addr, int prefix);
void bcache_do_one(char *haddr, char *caddr, int cbt, int type);
void bulist_do_one(char *cnaddr, char *haddr, char *coaddr, long int exp, 
		   int seq, int sta, int del, int mdl, long int cbs);
void stats_do_one(char *stat, int value);

void set_access(char *rule);
void set_debuglevel(char *level);
void set_tunnel_sitelocal(char *opt);
int set_key_md5 (char *key);
int set_key_sha1 (char *key);
int set_homeaddress(char *homeaddr);
int set_homeagent(char *homeagent);
void acl_file(char *path);
void acl_add_rule(char *policy, struct in6_addr addr, int prefix);

int check_uid(int uid);
int format_addr(char *addr, char *faddr);

int main (int argc, char **argv)
{
	int c;
	
	while (1)
	{
		int option_index = 0;
		static struct option long_options[] =
		{
			{"access", 0, 0, 'a'},
			{"rule", 1, 0, 'r'},
			{"rulelist", 1, 0, 'R'},
			{"statistics", 0, 0, 's'},
			{"bulist", 0, 0, 'l'},
			{"bcache", 0, 0, 'c'},
			{"help", 0, 0, '?'},
			{"version", 0, 0, 'V'},
			{"debuglevel", 1, 0, 'd'},
			{"tunnel_sitelocal", 1, 0, 't'},
			{"homeagent", 1, 0, 'H'},
			{"homeaddress", 1, 0, 'h'},
			{"MD5 key", 1, 0, 'M'},
			{"SHA-1 key", 1, 0, 'S'},
			{0, 0, 0, 0}
		};
		
		c = getopt_long (argc, argv, "M:S:R:H:acd:h:lr:t:k:sV?",
				 long_options, &option_index);
		if (c == -1)
			break;
		
		switch (c)
		{
		case 0:
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
			break;
		case 'r':
			set_access(optarg);
			break;
		case 'H':
			set_homeagent(optarg);
			break;
		case 'a':
			print_access();
			break;
		case 'c':
			print_bcache();
			break;
		case 'd':
			set_debuglevel(optarg);
			break;
		case 'R':
			acl_file(optarg);
			break;
		case 'M':
			set_key_md5 (optarg);
			break;
		case 'S':
			set_key_sha1 (optarg);
			break;
		case 'h':
			set_homeaddress(optarg);
			break;
		case 't':
			set_tunnel_sitelocal(optarg);
			break;
		case 's':
			print_stats();
			break;
		case 'l':
			print_bulist();
			break;
		case 'V':
			version();
			break;
		case '?':
			usage();
			break;
			
		default:
			printf ("mipdiag: invalid option -- %c. Try -? for help.\n", c);
		}
	}
	
	if (optind < argc)
	{
		printf ("non-option ARGV-elements: ");
		while (optind < argc)
			printf ("%s ", argv[optind++]);
		printf ("\n");
	}
	
	exit (0);
}

void usage()
{
	fprintf (stderr, "Usage: mipdiag [-acls]           Show Mobile IP information\n");
	fprintf (stderr, "       mipdiag [-hHdrtRMS value] Modify Mobile IP runtime parameters\n\n");
	fprintf (stderr, "       mipdiag [-?|--help]       Detailed usage syntax\n");
	fprintf (stderr, "       mipdiag [-V|--version]    Display version/author and exit\n\n");
	fprintf (stderr, "        -a, --access             show access control list\n");
	fprintf (stderr, "        -c, --bcache             show binding cache\n");
	fprintf (stderr, "        -l, --bulist             show bingding update list\n");
	fprintf (stderr, "        -s, --statistics         show statistics\n");
	fprintf (stderr, "        -M, --MD5-key            set MD5 key\n");
	fprintf (stderr, "        -S, --SHA-1-key          set SHA-1 key\n");
	fprintf (stderr, "        -h, --homeaddress        set MN's home address\n");
	fprintf (stderr, "        -H, --homeagent          set MN's home agent's address\n");
	fprintf (stderr, "        -d, --debuglevel         set module's debug level\n");
	fprintf (stderr, "        -r, --rule               set access control rule\n");
	fprintf (stderr, "        -R  --rulelist           read access control list from file\n");
	fprintf (stderr, "        -t, --tunnel_sitelocal   set tunnel sitelocal option\n");
}
void version()
{
	fprintf (stderr, "MIPL Mobile IPv6 0.5\n");
	fprintf (stderr, "mipdiag 0.10 (2000-03-20) Antti Tuominen\n");
}

void print_stats()
{
	FILE *stats;
	char stat[40];
	int value = 0;

	stats = fopen("/proc/net/mip6_stat", "r");
	if (stats != NULL) {
		printf ("Mobile IPv6 Statistics\n");
		while (!feof(stats)) {
			memset(stat, 0, 40);
			fscanf(stats, "%s = %d\n", stat, &value);
			stats_do_one(stat, value);
		}
		fclose(stats);
	} else {
		fprintf(stderr, "Could not read statistics information.\n");
		exit(ERR_NO_PROC);
	}
}

void stats_do_one(char *stat, int value)
{
	printf("%-25s : %d\n", stat, value);
}

void print_access()
{
	FILE *access;
	char policy, addr[33];
	int prefix;

	access = fopen("/proc/net/mip6_access", "r");
	if (access != NULL) {
		printf("Mobile IPv6 Mobile Node Access Control List\n");
		printf("Policy   Address range                               Prefix length\n");
		while (!feof(access)) {
			memset(addr, 0, 33);
			fscanf(access, "%c %32s %2x\n", &policy, addr, &prefix);
			access_do_one(policy, addr, prefix);
		}
		fclose(access);
	} else {
		fprintf(stderr, "Could not read access control list.\n");
		exit(ERR_NO_PROC);
	}
}

void access_do_one(char policy, char *addr, int prefix)
{
	char saddr[44];

	if (strlen(addr)) {
		memset(saddr, 0, 44);
		format_addr(addr, saddr);
		printf("%-8s %-44s %d\n", policy == '+' ? "ALLOW" : "DENY", saddr, prefix);
	}
}

void print_bcache()
{
	FILE *bcache;
	char haddr[33], caddr[33];
	long int callback_time = 0;
	int type = 0;

	bcache = fopen("/proc/net/mip6_bcache", "r");
	if (bcache != NULL) {
		printf("Mobile IPv6 Binding cache\n");
		printf("Home Address                                Care-of Address                             Lifetime  Type\n");
		while (!feof(bcache)) {
			memset(haddr, 0, 33);
			memset(caddr, 0, 33);
			fscanf(bcache, "h=%32s  c=%32s (e=%ld,t=%d)\n",
			       haddr, caddr, &callback_time, &type);
			bcache_do_one(haddr, caddr, callback_time, type);
		}
		fclose(bcache);
	} else {
		fprintf(stderr, "Could not read binding cache information.\n");
		exit(ERR_NO_PROC);
	}
}

void bcache_do_one(char *haddr, char *caddr, int cbt, int type)
{
	char buf1[44], buf2[44];
	if (strlen(haddr) && strlen(caddr)) {
		memset(buf1,0,44);
		memset(buf2,0,44);

		format_addr(haddr, buf1);
		format_addr(caddr, buf2);
		printf("%-44s%-44s%-10d%-5d\n", buf1, buf2, cbt, type);
	}
}

void print_bulist()
{
	FILE *bulist;
	char cnaddr[33], haddr[33], coaddr[33];
	int seq = 0, sta = 0, del = 0, mdl = 0;
	long int exp = 0, cbs = 0;

	bulist = fopen("/proc/net/mip6_bul", "r");
	if (bulist != NULL) {
		printf ("Mobile IPv6 Binding update list\n");
		while (!feof(bulist)) {
			memset(cnaddr, 0, 33);
			memset(haddr, 0, 33);
			memset(coaddr, 0, 33);
			fscanf(bulist, "cna=%32s ha=%32s coa=%32s\nexp=%ld seq=%d sta=%d del=%d mdl=%d cbs=%ld\n",
			       cnaddr, haddr, coaddr, &exp, &seq, &sta, &del, &mdl, &cbs);
			bulist_do_one(cnaddr, haddr, coaddr, exp, seq, sta, del, mdl, cbs);
		}
		fclose(bulist);
	} else {
		fprintf(stderr, "Could not read binding update list information.\n");
		exit(ERR_NO_PROC);
	}
}

void bulist_do_one(
	char *cnaddr, char *haddr, char *coaddr, 
	long int exp, int seq, int sta, int del, int mdl, long int cbs)
{

	char buf1[44], buf2[44], buf3[44];
	
	if (strlen(cnaddr) && strlen(haddr) && strlen(coaddr)) {
		memset(buf1,0,44);
		memset(buf2,0,44);
		memset(buf3,0,44);

		format_addr(cnaddr, buf1);
		format_addr(haddr, buf2);
		format_addr(coaddr, buf3);
		printf("Recipient CN: %s\n", buf1);
		printf("BINDING home address: %s care-of address: %s\n", buf2, buf3);
		printf("        expires: %ld sequence: %d state: %d\n", exp, seq, sta);
		printf("        delay: %d max delay %d callback time: %ld\n\n", del, mdl, cbs);
	}
}

void acl_add_rule(char *policy, struct in6_addr addr, int prefix)
{
	FILE *acl;
	char cpolicy;

	if (!strcasecmp(policy, "DENY"))
		cpolicy = '-';
	else if (!strcasecmp(policy, "ALLOW"))
		cpolicy = '+';
	else if (!strcasecmp(policy, "FLUSH"))
		cpolicy = '*';
	else {
		fprintf(stderr, "Access list format error: unknown policy \"%s\".\n",
			policy);
		exit(1);
	}

	acl = fopen("/proc/sys/net/ipv6/mobility/mobile_node_list", "r+");
	if (acl != NULL) {
		if (cpolicy == '*')
			fprintf(acl, "*");
		else
			fprintf(acl, "%c %08x%08x%08x%08x %02x", cpolicy,
				htonl(addr.s6_addr32[0]), 
				htonl(addr.s6_addr32[1]),
				htonl(addr.s6_addr32[2]), 
				htonl(addr.s6_addr32[3]),
				prefix);
		fclose(acl);
	}

}

void acl_file(char *path)
{
	FILE *acl;
	char buffer[128];

	if (!strcmp(path, "-"))
		acl = stdin;
	else
		acl = fopen(path, "r");

	if (acl != NULL) {
		while (!feof(acl)) {
			if (fgets(buffer, 128, acl) != NULL)
				set_access(buffer);
		}
		fclose(acl);
	}
	
}

void set_access(char *rule) 
{
	int plen = 0;
	char *policy, *addr, *n, *p;
	struct in6_addr saddr;

	if (!check_uid(UID0)) {
		fprintf(stderr, "You must be root (uid = 0) to do this.\n");
		exit(ERR_UID);
	}

	if (strchr(rule, '#')) {
		return;
	}
	policy = (char *) malloc(128 * sizeof(char));
	addr = (char *) malloc(128 * sizeof(char));

	sscanf(rule, "%s %s", policy, addr); 
	
	if (strncasecmp(policy, "FLUSH", 5)) {
		n = strchr(addr, '/');
		if (n) {
			p = n + 1;
			*n = '\0';
			plen = strtoul(p, &p, 10);
			if (*p) exit(1);
		} else {
			fprintf(stderr, "Access list format error: no prefix length.\n");
			exit(1);
		}
		
		if(!inet_pton(AF_INET6, addr, &saddr)) {
			fprintf(stderr, "Access list format error: invalid IPv6 address.\n");
			exit(1);
		}
	} 
	acl_add_rule(policy, saddr, plen);
}

void set_debuglevel(char *level)
{
	FILE *debug;
	int ilevel = atoi(level);

	if (!check_uid(UID0)) {
		fprintf(stderr, "You must be root (uid = 0) to do this.\n");
		exit(ERR_UID);
	}
	debug = fopen("/proc/sys/net/ipv6/mobility/debuglevel", "r+");
	if (debug != NULL) {
		if (fprintf(debug, "%d", ilevel) > 0)
			printf("Debug level set to %d\n", ilevel);
		else
			fprintf(stderr, "Unable to set debug level.\n");
		fclose(debug);
	} else {
		fprintf(stderr, "Could not set debug level. Is module loaded?\n");
		exit(ERR_NO_SYSCTL);
	}
}

void set_tunnel_sitelocal(char *opt)
{
	FILE *tunnel_sitelocal;
	int iopt = -1;

	if (!strcasecmp(opt, "YES") || !strcasecmp(opt, "ON") || 
	    !strcasecmp(opt, "Y") || !strcmp(opt, "1"))
		iopt = 1;
	else if (!strcasecmp(opt, "NO") || !strcasecmp(opt, "OFF") ||
		 !strcasecmp(opt, "N") || !strcmp(opt, "0"))
		iopt = 0;
	else {
		fprintf(stderr, "Invalid value for option: %s\n", opt);
		exit(ERR_ARG);
	}

	if (!check_uid(UID0)) {
		fprintf(stderr, "You must be root (uid = 0) to do this.\n");
		exit(ERR_UID);
	}

	tunnel_sitelocal = fopen("/proc/sys/net/ipv6/mobility/tunnel_sitelocal", "r+");
	if (tunnel_sitelocal != NULL) {
		if (fprintf(tunnel_sitelocal, "%d", iopt) > 0)
			if (iopt==0)
				printf("Tunnel sitelocal option set to 'no'.\n");
			else
				printf("Tunnel sitelocal option set to 'yes'.\n");
		else
			fprintf(stderr, "Unable to set tunnel sitelocal option.\n");
		fclose(tunnel_sitelocal);
	} else {
		fprintf(stderr, "Could not set tunnel sitelocal option. Is module loaded?\n");
		exit(ERR_NO_SYSCTL);
	}
}

int
base64_to_bin (uint8_t *dst,char *src,int size);

#define __set_key_md5(key)\
        __set_key (key, 128)
#define __set_key_sha1(key)\
        __set_key (key, 160)

int
__set_key (char *key, int keylen);

int
set_key_md5 (char * key)
{
	if (key)
		return __set_key_md5 (key);
	else
		return -1;
}

int
set_key_sha1 (char * key)
{
	if (key)
		return __set_key_sha1 (key);
	else
		return -1;
}

int 
__set_key (char *key, int keylen)
{
	FILE *keyf;
	uint32_t * k;
	int ret = 0;
	char cn [10];
	char fn [] = "/proc/sys/net/ipv6/mobility/key";
	char * fncn;
	unsigned char * os;
	int osl = 0;
	int i = 0;

	if (keylen <= 0
	    || (keylen % 4)
	    || (keylen % 8))
		return -1;
	k = (uint32_t *) malloc (sizeof (uint32_t) * ((keylen / 8) * 2));
	if (!k)
		return -1;
	ret = base64_to_bin ((uint8_t *) k, key, keylen / 8);
	if (ret == -EINVDIGIT)
	{
		printf("An invalid digit was used in the key\n");
		return -EINVDIGIT;
	}
	else if (ret == -ESRCTOOLONG)
	{
		printf("The key was longer than expected\n");
		return -ESRCTOOLONG;
	}		
	else if (ret == ESRCTOOSHORT)
	{
		printf("The key was shorter than expected\n");
		return -ESRCTOOSHORT;
	}				
	if (!check_uid(UID0)) {
		fprintf(stderr, "You must be root (uid = 0) to do this.\n");
		exit(ERR_UID);
	}
	sprintf (cn, "%d", keylen);
	fncn = (char *) malloc (strlen (fn) + strlen (cn) + 1);
	if (!fncn)
		return -1;
	sprintf (fncn, "%s%s", fn, cn);
	keyf = fopen (fncn, "r+");
	if (keyf != NULL) {
		osl = (keylen / 8) * 2 + 1;
		os = (char *) malloc (osl);
		if (!os)
			return -1;
		for (i = 0; i < (keylen / 8) / 4; i++)
			sprintf (&os [i * 8], "%08x", k [i]);
		os [osl - 1] = '\0';
		fprintf (keyf, "%s", os);
		printf ("Setting key to: %s\n", os);
		fclose(keyf);
	} else {
		fprintf(stderr, "Could not set key.\n");
		exit(ERR_NO_SYSCTL);
	}
	free (k);
	free (os);
	free (fncn);
	return 0;
}
int set_homeaddress(char *addr)
{
	FILE *homeaddress;
	struct in6_addr ha_addr;
	char *p, *n;
	int prefix = 128;

	if (!check_uid(UID0)) {
		fprintf(stderr, "You must be root (uid = 0) to do this.\n");
		exit(ERR_UID);
	}

	n = strchr(addr, '/');
	if (n) {
		p = n + 1;
		*n = '\0';
		prefix = strtoul(p, &p, 10);
		if (*p) prefix = 128;
	}

	inet_pton(AF_INET6, addr, &ha_addr);

	homeaddress = fopen("/proc/sys/net/ipv6/mobility/home_address", "r+");
	if (homeaddress != NULL) {
		if (fprintf(homeaddress, "%08x%08x%08x%08x %02x", 
			    (int)ntohl(ha_addr.s6_addr32[0]),
			    (int)ntohl(ha_addr.s6_addr32[1]),
			    (int)ntohl(ha_addr.s6_addr32[2]),
			    (int)ntohl(ha_addr.s6_addr32[3]),
			    prefix) > 0)
			printf("Home address set to %s.\n", addr);
		else
			fprintf(stderr, "Unable to set home address.\n");
		fclose(homeaddress);
	} else {
		fprintf(stderr, "Could not set home address. Is module loaded?\n");
		exit(ERR_NO_SYSCTL);
	}

	return 0;
}

int set_homeagent(char *addr)
{
	FILE *homeagent;
	struct in6_addr ha_addr;
	char *p, *n;
	int prefix = 128;

	if (!check_uid(UID0)) {
		fprintf(stderr, "You must be root (uid = 0) to do this.\n");
		exit(ERR_UID);
	}

	n = strchr(addr, '/');
	if (n) {
		p = n + 1;
		*n = '\0';
		prefix = strtoul(p, &p, 10);
		if (*p) prefix = 128;
	}

	inet_pton(AF_INET6, addr, &ha_addr);

	homeagent = fopen("/proc/sys/net/ipv6/mobility/home_agent_address", "r+");
	if (homeagent != NULL) {
		if (fprintf(homeagent, "%08x%08x%08x%08x", (int)ntohl(ha_addr.s6_addr32[0]),
			    (int)ntohl(ha_addr.s6_addr32[1]),
			    (int)ntohl(ha_addr.s6_addr32[2]),
			    (int)ntohl(ha_addr.s6_addr32[3])) > 0)
			printf("Home agent address set to %s.\n", addr);
		else
			fprintf(stderr, "Unable to set home agent address.\n");
		fclose(homeagent);
	} else {
		fprintf(stderr, "Could not set home agent address. Is module loaded?\n");
		exit(ERR_NO_SYSCTL);
	}

	return 0;
}

int check_uid(int uid)
{
	if (geteuid() != uid) return 0;
	else return 1;
}

int format_addr(char *addr, char *faddr)
{
	int taddr[8];
	int i, mz = 0, tz = 0, zl = -1;
	char *p;

	sscanf(addr, "%04x%04x%04x%04x%04x%04x%04x%04x",
	       &taddr[0], &taddr[1], &taddr[2], &taddr[3],
	       &taddr[4], &taddr[5], &taddr[6], &taddr[7]);

	for (i = 0; i < 8; i++) {
		if (taddr[i]==0) tz++;
		else {
			if ((tz > 0) && (tz > mz)) {
				zl = i - tz;
				mz = tz;
			}
			tz = 0;
		}
	}
	p = faddr;
	for (i = 0; i < 8; i++) {
		if (i == zl) {
			i += mz - 1;
			p += sprintf(p, ":");
			if ((zl + mz) == 8)
				p += sprintf(p, ":");
			continue;
		}
		if (i != 0)
			p += sprintf(p, ":");
		p += sprintf(p, "%x", taddr[i]);
	}
	p = NULL;
	return strlen(faddr);
}

/*
 * dst: pointer to destination buffer
 * src: pointer to base64 string
 * size: size (in BYTES) of destination buffer
 */
int base64_to_bin(uint8_t *dst,char *src,int size)
{
  int i,j;
  char *p;
  int offset,count;
  uint32_t digitval;
  uint8_t buf[4];

  offset = 18;
  *(uint32_t *)buf = 0;
  count = size;
  memset(dst,0,size);
  
  for (i = 0; src[i] != '\0'; i++) {
    p = strchr(base64_digits,src[i]);
    if (p == NULL)
      return (-EINVDIGIT);
    digitval = p - base64_digits;
    
    *(uint32_t *)buf |= htonl(digitval << offset);
    
    if (offset == 0) {
      for (j = 0; (j < 3) && (count > 0); j++,count--)
        dst[j] = buf[j + 1];

      offset = 18;
      dst += 3;
      *(uint32_t *)buf = 0;
    }
    else
      offset -= 6;
  }
  
  if (offset != 18) 
    for (j = 0; (j < 3) && (count > 0); j++,count--)
      dst[j] = buf[j + 1];
  
  if (i < ((8*size + 5) / 6))
    return (-ESRCTOOSHORT);
  else if (i > ((8*size + 5) / 6))
    return (-ESRCTOOLONG);

  return (0);
}
